import unittest


class SwiftFacilityTest(unittest.TestCase):

    def test_something(self):
        expected = True
        actual = True
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()
